# Copyright (c) OpenMMLab. All rights reserved.
import torch
from mmengine.dataset import DefaultSampler
from mmengine.hooks import (CheckpointHook, DistSamplerSeedHook, IterTimerHook,
                            LoggerHook, ParamSchedulerHook)
from mmengine.optim import AmpOptimWrapper, CosineAnnealingLR, LinearLR
from peft import LoraConfig
from torch.optim import AdamW
from transformers import (AutoModelForCausalLM, WhisperModel, AutoTokenizer,
                          BitsAndBytesConfig, WhisperProcessor)

from xtuner.dataset import LLaMASpeechDataset
from xtuner.dataset.collate_fns import default_collate_fn
from xtuner.dataset.map_fns import para_map_fn, template_map_fn_factory, para_test_map_fn
from xtuner.dataset.samplers import LengthGroupedSampler
from xtuner.engine.hooks import DatasetInfoHook, EvaluateChatHook
from xtuner.engine.runner import TrainLoop, TestLoop
from xtuner.model import LLaMASpeechModel
from xtuner.utils import PROMPT_TEMPLATE
from xtuner.evaluation.metrics.bleu import bleu
from xtuner.evaluation.metrics.rouge import rouge
from xtuner.evaluation.metrics.bert import bert
from xtuner.evaluation.metrics.loss import loss
from os import environ

#######################################################################
#                          PART 1  Settings                           #
#######################################################################
# Model
llm_name_or_path = 'meta-llama/Llama-2-7b-chat-hf'
speech_encoder_name_or_path = 'openai/whisper-large-v3'

# Data
train_name = environ.get('TRAIN_NAME', 'train')
test_name = environ.get('TEST_NAME', 'test_emo')
post_fix = environ.get('POST_FIX', '')
resume_ckp = environ.get('RESUME', 'False')
exp_path = environ.get('EXP_PATH', '')
train_data_path = f'{train_name}.json'
test_data_path = f'{test_name}.json'

prompt_template = PROMPT_TEMPLATE.llama2_chat
# prompt_template = None
max_length = int(2048 - (336 / 14)**2)

# Scheduler & Optimizer
batch_size = 16  # per_device
accumulative_counts = 2
dataloader_num_workers = 0
max_epochs = 2
optim_type = AdamW
lr = 2e-4
betas = (0.9, 0.999)
weight_decay = 0
max_norm = 1  # grad clip
warmup_ratio = 0.03

# Save
save_steps = 500
save_total_limit = 2  # Maximum checkpoints to keep (-1 means unlimited)

# Evaluate the generation performance during the training
evaluation_freq = 500
SYSTEM_LS = ''
evaluation_speech = 'RAVDESS-Speech/Actor_01/03-01-05-02-01-01-01.wav' # Kids are talking by the door.
evaluation_inputs = ['']

#######################################################################
#            PART 2  Model & Tokenizer & Image Processor              #
#######################################################################
tokenizer = dict(
    type=AutoTokenizer.from_pretrained,
    pretrained_model_name_or_path=llm_name_or_path,
    trust_remote_code=True,
    padding_side='right')

speech_processor = dict(
    type=WhisperProcessor.from_pretrained,
    pretrained_model_name_or_path=speech_encoder_name_or_path,
    trust_remote_code=True)

model = dict(
    type=LLaMASpeechModel,
    freeze_llm=True,
    freeze_speech_encoder=True,
    llm=dict(
        type=AutoModelForCausalLM.from_pretrained,
        pretrained_model_name_or_path=llm_name_or_path,
        trust_remote_code=True,
        torch_dtype=torch.float16,
        quantization_config=dict(
            type=BitsAndBytesConfig,
            load_in_4bit=True,
            load_in_8bit=False,
            llm_int8_threshold=6.0,
            llm_int8_has_fp16_weight=False,
            bnb_4bit_compute_dtype=torch.float16,
            bnb_4bit_use_double_quant=True,
            bnb_4bit_quant_type='nf4')),
    llm_lora=dict(
        type=LoraConfig,
        r=512,
        lora_alpha=256,
        lora_dropout=0.05,
        bias='none',
        task_type='CAUSAL_LM'),
    pooling=2,
    speech_encoder=dict(
        type=WhisperModel.from_pretrained,
        pretrained_model_name_or_path=speech_encoder_name_or_path),
    speech_encoder_lora=dict(
        type=LoraConfig, r=64, lora_alpha=16, lora_dropout=0.05, bias='none'))

#######################################################################
#                      PART 3  Dataset & Dataloader                   #
#######################################################################
llama_speech_dataset = dict(
    type=LLaMASpeechDataset,
    data_path=train_data_path,
    tokenizer=tokenizer,
    speech_processor=speech_processor,
    dataset_map_fn=para_map_fn,
    split='train',
    template_map_fn=dict(
        type=template_map_fn_factory, template=prompt_template),
    max_length=max_length)

train_dataloader = dict(
    batch_size=batch_size,
    num_workers=dataloader_num_workers,
    dataset=llama_speech_dataset,
    sampler=dict(
        type=LengthGroupedSampler,
        length_property='modality_length',
        per_device_batch_size=batch_size * accumulative_counts),
    collate_fn=dict(type=default_collate_fn))

llama_speech_test_dataset = dict(
    type=LLaMASpeechDataset,
    data_path=test_data_path,
    tokenizer=tokenizer,
    speech_processor=speech_processor,
    dataset_map_fn=para_test_map_fn,
    split=test_name,
    template_map_fn=dict(
        type=template_map_fn_factory, template=prompt_template),
    max_length=max_length)

# Only support batch size 1
test_dataloader = dict(
    batch_size=1,
    num_workers=dataloader_num_workers,
    dataset=llama_speech_test_dataset,
    sampler=dict(type=DefaultSampler, shuffle=False),
    collate_fn=dict(type=default_collate_fn))

test_evaluator = [
    dict(type=bleu, prefix='bleu'),
    dict(type=bert, prefix='bert'),
    dict(type=rouge, prefix='rouge'),
]

#######################################################################
#                    PART 4  Scheduler & Optimizer                    #
#######################################################################
# optimizer
optim_wrapper = dict(
    type=AmpOptimWrapper,
    optimizer=dict(
        type=optim_type, lr=lr, betas=betas, weight_decay=weight_decay),
    clip_grad=dict(max_norm=max_norm, error_if_nonfinite=False),
    accumulative_counts=accumulative_counts,
    loss_scale='dynamic',
    dtype='float16')

# learning policy
# More information: https://github.com/open-mmlab/mmengine/blob/main/docs/en/tutorials/param_scheduler.md  # noqa: E501
param_scheduler = [
    dict(
        type=LinearLR,
        start_factor=1e-5,
        by_epoch=True,
        begin=0,
        end=warmup_ratio * max_epochs,
        convert_to_iter_based=True),
    dict(
        type=CosineAnnealingLR,
        eta_min=0.0,
        by_epoch=True,
        begin=warmup_ratio * max_epochs,
        end=max_epochs,
        convert_to_iter_based=True)
]

# train, val, test setting
train_cfg = dict(type=TrainLoop, max_epochs=max_epochs, val_interval=save_steps)
test_cfg = dict(
    type=TestLoop,
    tokenizer=tokenizer,
    work_dir=f'{exp_path}/exp/vs_llm{post_fix}',
    task_name='conv'
)

#######################################################################
#                           PART 5  Runtime                           #
#######################################################################
# Log the dialogue periodically during the training process, optional
custom_hooks = [
    dict(type=DatasetInfoHook, tokenizer=tokenizer),
    dict(
        type=EvaluateChatHook,
        tokenizer=tokenizer,
        speech_processor=speech_processor,
        every_n_iters=evaluation_freq,
        evaluation_inputs=evaluation_inputs,
        evaluation_speech=evaluation_speech,
        system=SYSTEM_LS,
        prompt_template=prompt_template)
]

# configure default hooks
default_hooks = dict(
    # record the time of every iteration.
    timer=dict(type=IterTimerHook),
    # print log every 10 iterations.
    logger=dict(type=LoggerHook, log_metric_by_epoch=False, interval=10),
    # enable the parameter scheduler.
    param_scheduler=dict(type=ParamSchedulerHook),
    # save checkpoint per `save_steps`.
    checkpoint=dict(
        type=CheckpointHook,
        by_epoch=False,
        interval=save_steps,
        max_keep_ckpts=save_total_limit),
    # set sampler seed in distributed evrionment.
    sampler_seed=dict(type=DistSamplerSeedHook),
)

# configure environment
env_cfg = dict(
    # whether to enable cudnn benchmark
    cudnn_benchmark=False,
    # set multi process parameters
    mp_cfg=dict(mp_start_method='fork', opencv_num_threads=0),
    # set distributed parameters
    dist_cfg=dict(backend='nccl'),
)

# set visualizer
visualizer = None

# set log level
log_level = 'INFO'

# load from which checkpoint
load_from = None

# whether to resume training from the loaded checkpoint
resume = eval(resume_ckp)

# Defaults to use random seed and disable `deterministic`
randomness = dict(seed=None, deterministic=False)

# set log processor
log_processor = dict(by_epoch=False)
